/*
 *  map.c
 *
 *  Copyright 2013 Michael Zillgith
 *
 *	This file is part of libIEC61850.
 *
 *	libIEC61850 is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	libIEC61850 is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with libIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 *	See COPYING file for the complete license text.
 */

#include "map.h"
#include <stdlib.h>

typedef struct sMapEntry {
	void* key;
	void* value;
} MapEntry;

static int
comparePointerKeys(void* key1, void* key2)
{
	return key2 - key1;
}

Map
Map_create()
{
	Map map = calloc(1, sizeof(struct sMap));
	map->entries = LinkedList_create();
	map->compareKeys = comparePointerKeys;
	return map;
}

int
Map_size(Map map)
{
	return LinkedList_size(map->entries);
}

void*
Map_addEntry(Map map, void* key, void* value)
{
	MapEntry* entry = malloc(sizeof(MapEntry));
	entry->key = key;
	entry->value = value;
	LinkedList_add(map->entries, entry);

	return entry->key;
}

void*
Map_removeEntry(Map map, void* key, bool deleteKey)
{
	LinkedList element = map->entries;
	LinkedList lastElement = element;
	MapEntry* entry;
	void* value = NULL;

	while ((element = LinkedList_getNext(element)) != NULL) {
		entry = (MapEntry*) element->data;

		if (map->compareKeys(key, entry->key) == 0) {
			lastElement->next = element->next;
			value = entry->value;

			if (deleteKey == true)
				free(entry->key);
			free(entry);
			free(element);

			break;
		}

		lastElement = element;
	}

	return value;
}

void*
Map_getEntry(Map map, void* key)
{
	LinkedList element = map->entries;
	MapEntry* entry;

	while ((element = LinkedList_getNext(element)) != NULL) {
		entry = (MapEntry*) element->data;
		if (map->compareKeys(key, entry->key) == 0) {
			return entry->value;
		};
	}

	return NULL;
}

void
Map_delete(Map map, bool deleteKey)
{
	LinkedList element = map->entries;
	MapEntry* entry;

	while ((element = LinkedList_getNext(element)) != NULL) {
		entry = (MapEntry*) element->data;
		if (deleteKey == true)
			free(entry->key);
		free(entry->value);
	}

	LinkedList_destroy(map->entries);
	free(map);
}

void
Map_deleteStatic(Map map, bool deleteKey)
{
	LinkedList element = map->entries;
	MapEntry* entry;

	if (deleteKey == true) {
		while ((element = LinkedList_getNext(element)) != NULL) {
			entry = (MapEntry*) element->data;
			free(entry->key);
		}
	}

	LinkedList_destroy(map->entries);
	free(map);
}

void
Map_deleteDeep(Map map, bool deleteKey, void (*valueDeleteFunction) (void*))
{
	LinkedList element = map->entries;
	MapEntry* entry;

	while ((element = LinkedList_getNext(element)) != NULL) {
		entry = (MapEntry*) element->data;
		if (deleteKey == true)
			free(entry->key);
		valueDeleteFunction(entry->value);
	}

	LinkedList_destroy(map->entries);
	free(map);
}
