/*
 *  mms_information_report.c
 *
 *  Copyright 2013 Michael Zillgith
 *
 *  This file is part of libIEC61850.
 *
 *  libIEC61850 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libIEC61850 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  See COPYING file for the complete license text.
 */

#include <stdlib.h>
#include <stdbool.h>

#include "mms_server_internal.h"
#include "mms_common_internal.h"

#if 1

static int
encodeAccessResult(MmsValue* value, uint8_t* buffer, int bufPos, bool encode);

static int
encodeArrayAccessResult(MmsValue* value, uint8_t* buffer, int bufPos, bool encode)
{
    int elementsSize = 0;

    int i;

    int arraySize = value->value.array.size;

    for (i = 0; i < arraySize; i++) {
        MmsValue* element = value->value.array.components[i];

        elementsSize += encodeAccessResult(element, NULL, 0, false);
    }

    if (encode) {
        buffer[bufPos++] = 0xa1; /* tag for array */
        bufPos = BerEncoder_encodeLength(elementsSize, buffer, bufPos);

        for (i = 0; i < arraySize; i++) {
            MmsValue* element = value->value.array.components[i];

            bufPos = encodeAccessResult(element, buffer, bufPos, true);
        }

        return bufPos;
    }
    else {
        int size = 1 + elementsSize + BerEncoder_determineLengthSize(elementsSize);

        return size;
    }
}

static int
encodeStructuredAccessResult(MmsValue* value, uint8_t* buffer, int bufPos, bool encode)
{
    int componentsSize = 0;

    int i;

    int components = value->value.structure.componentCount;

    for (i = 0; i < components; i++) {
        MmsValue* component = value->value.structure.components[i];

        componentsSize += encodeAccessResult(component, NULL, 0, false);
    }

    if (encode) {
        buffer[bufPos++] = 0xa2; /* tag for structure */
        bufPos = BerEncoder_encodeLength(componentsSize, buffer, bufPos);

        for (i = 0; i < components; i++) {
            MmsValue* component = value->value.structure.components[i];

            bufPos = encodeAccessResult(component, buffer, bufPos, true);
        }

        return bufPos;
    }
    else {
        int size = 1 + componentsSize + BerEncoder_determineLengthSize(componentsSize);

        return size;
    }
}

static int
encodeAccessResult(MmsValue* value, uint8_t* buffer, int bufPos, bool encode)
{
    int size;

    switch (value->type) {
    case MMS_STRUCTURE:
        if (encode)
            bufPos = encodeStructuredAccessResult(value, buffer, bufPos, true);
        else
            size = encodeStructuredAccessResult(value, buffer, bufPos, false);

        break;

    case MMS_ARRAY:
        if (encode)
            bufPos = encodeArrayAccessResult(value, buffer, bufPos, true);
        else
            size = encodeArrayAccessResult(value, buffer, bufPos, false);
        break;

    case MMS_DATA_ACCESS_ERROR:
        if (encode)
            bufPos = BerEncoder_encodeUInt32(value->value.dataAccessError.code, buffer, bufPos);
        else
            size = 2 + BerEncoder_UInt32determineEncodedSize(value->value.dataAccessError.code);
        break;

    case MMS_VISIBLE_STRING:
        if (encode)
            bufPos = BerEncoder_encodeStringWithTag(0x8a, value->value.visibleString, buffer, bufPos);
        else
            size = BerEncoder_determineEncodedStringSize(value->value.visibleString);
        break;
    case MMS_UNSIGNED:
        if (encode)
            bufPos = BerEncoder_encodeAsn1PrimitiveValue(0x86, value->value.unsignedInteger, buffer, bufPos);
        else
            size = 2 + value->value.unsignedInteger->size;
        break;
    case MMS_INTEGER:
        if (encode)
            bufPos = BerEncoder_encodeAsn1PrimitiveValue(0x85, value->value.integer, buffer, bufPos);
        else
            size = 2 + value->value.integer->size;
        break;
    case MMS_UTC_TIME:
        if (encode)
            bufPos = BerEncoder_encodeOctetString(0x91, value->value.utcTime, 8, buffer, bufPos);
        else
            size = 10;
        break;
    case MMS_BIT_STRING:
        if (encode)
            bufPos = BerEncoder_encodeBitString(0x84, value->value.bitString.size,
                    value->value.bitString.buf, buffer, bufPos);
        else
            size = BerEncoder_determineEncodedBitStringSize(value->value.bitString.size);
        break;
    case MMS_BOOLEAN:
        if (encode)
            bufPos = BerEncoder_encodeBoolean(0x83, value->value.boolean, buffer, bufPos);
        else
            size = 3;
        break;
    case MMS_BINARY_TIME:
        if (encode)
            bufPos = BerEncoder_encodeOctetString(0x8c, value->value.binaryTime.buf,
                    value->value.binaryTime.size, buffer, bufPos);
        else
            size = 2 + value->value.binaryTime.size;
        break;
    case MMS_OCTET_STRING:
        if (encode)
            bufPos = BerEncoder_encodeOctetString(0x89, value->value.octetString.buf,
                    value->value.octetString.size, buffer, bufPos);
        else
            size = 1 + BerEncoder_determineLengthSize(value->value.octetString.size)
                        + value->value.octetString.size;
        break;

    case MMS_FLOAT:
        {
            int floatSize = (value->value.floatingPoint.formatWidth / 8) + 1;

            if (encode) {
                bufPos = BerEncoder_encodeTL(0x87, floatSize, buffer, bufPos);
                bufPos = BerEncoder_encodeFloat(value->value.floatingPoint.buf,
                        value->value.floatingPoint.formatWidth,
                        value->value.floatingPoint.exponentWidth,
                        buffer, bufPos);
            }
            else
                size = floatSize + 2; /* 2 for tag and length */
        }
        break;
    case MMS_STRING:
        if (encode)
            bufPos = BerEncoder_encodeStringWithTag(0x90, value->value.mmsString, buffer, bufPos);
        else
            size = BerEncoder_determineEncodedStringSize(value->value.mmsString);
        break;
    default:
        printf("encodeAccessResult: error unsupported type!\n");
        break;
    }

    if (encode)
        return bufPos;
    else
        return size;
}

void
MmsServerConnection_sendInformationReport(MmsServerConnection* self, char* domainId, char* itemId, LinkedList values)
{

    uint32_t variableAccessSpecSize = 0;
    uint32_t objectNameSize = 0;

    uint32_t accessResultSize = 0;
    uint32_t listOfAccessResultSize;
    uint32_t informationReportSize;

    if (domainId != NULL) {
//        report->variableAccessSpecification.choice.variableListName.present = ObjectName_PR_domainspecific;
//        report->variableAccessSpecification.choice.variableListName.choice.domainspecific.domainId.size = strlen(domainId);
//        report->variableAccessSpecification.choice.variableListName.choice.domainspecific.domainId.buf = domainId;
//        report->variableAccessSpecification.choice.variableListName.choice.domainspecific.itemId.size = strlen(itemId);
//        report->variableAccessSpecification.choice.variableListName.choice.domainspecific.itemId.buf = itemId;
    }
    else {
        objectNameSize = BerEncoder_determineEncodedStringSize(itemId);
        variableAccessSpecSize += objectNameSize;
        variableAccessSpecSize += BerEncoder_determineLengthSize(objectNameSize);
        variableAccessSpecSize += 1; /* space for tag (a1) */
    }

    int variableCount = LinkedList_size(values);

    /* iterate values list and add values to the accessResultList */
    LinkedList value = LinkedList_getNext(values);

    int i;

    for (i = 0; i < variableCount; i++) {

       MmsValue* data = (MmsValue*) value->data;

       accessResultSize += encodeAccessResult(data, NULL, 0, false);

        value = LinkedList_getNext(value);
    }

    listOfAccessResultSize = accessResultSize +
            BerEncoder_determineLengthSize(accessResultSize) + 1;

    uint32_t informationReportContentSize = variableAccessSpecSize + listOfAccessResultSize;

    informationReportSize = 1 +  informationReportContentSize +
            BerEncoder_determineLengthSize(informationReportContentSize);

    if (DEBUG) printf("sendInfReport: %i items\n", variableCount);

    ByteBuffer* reportBuffer = ByteBuffer_create(NULL, self->maxPduSize);

    uint8_t* buffer = reportBuffer->buffer;
    int bufPos = 0;


    /* encode */
    bufPos = BerEncoder_encodeTL(0xa3, informationReportSize, buffer, bufPos);
    bufPos = BerEncoder_encodeTL(0xa0, informationReportContentSize, buffer, bufPos);

    bufPos = BerEncoder_encodeTL(0xa1, objectNameSize, buffer, bufPos);
    bufPos = BerEncoder_encodeStringWithTag(0x80, itemId, buffer, bufPos);

    bufPos = BerEncoder_encodeTL(0xa0, accessResultSize, buffer, bufPos);

    value = LinkedList_getNext(values);

    for (i = 0; i < variableCount; i++) {

        MmsValue* data = (MmsValue*) value->data;

        bufPos = encodeAccessResult(data, buffer, bufPos, true);

        value = LinkedList_getNext(value);
    }

    if (DEBUG) printf("sendInfReport: %i encoded\n", bufPos);

    reportBuffer->size = bufPos;

    IsoConnection_sendMessage(self->isoConnection, reportBuffer);

    ByteBuffer_destroy(reportBuffer);
}

#endif

#if 0

static void
freeMmsInformationReportPdu(MmsPdu_t* mmsPdu)
{
    int accessResultSize =
            mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.listOfAccessResult.list.count;

    AccessResult_t** accessResult =
            mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.listOfAccessResult.list.array;

    mmsMsg_deleteAccessResultList(accessResult, accessResultSize);

    if (mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.variableAccessSpecification.
            choice.variableListName.present == ObjectName_PR_domainspecific)
    {
        mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.
            variableAccessSpecification.choice.variableListName.choice.domainspecific.domainId.size = 0;
        mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.
            variableAccessSpecification.choice.variableListName.choice.domainspecific.domainId.buf = NULL;

        mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.
            variableAccessSpecification.choice.variableListName.choice.domainspecific.itemId.size = 0;
        mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.
            variableAccessSpecification.choice.variableListName.choice.domainspecific.itemId.buf = NULL;
    }
    else {
        mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.
            variableAccessSpecification.choice.variableListName.choice.vmdspecific.size = 0;
        mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.
            variableAccessSpecification.choice.variableListName.choice.vmdspecific.buf = NULL;
    }

    mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.listOfAccessResult.list.array = 0;
    mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.listOfAccessResult.list.count = 0;
    mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport.listOfAccessResult.list.size = 0;

    asn_DEF_MmsPdu.free_struct(&asn_DEF_MmsPdu, mmsPdu, 0);
}

void
MmsServerConnection_sendInformationReport(MmsServerConnection* self, char* domainId, char* itemId, LinkedList values)
{
	MmsPdu_t* mmsPdu = calloc(1, sizeof(MmsPdu_t));

	mmsPdu->present = MmsPdu_PR_unconfirmedPDU;
	mmsPdu->choice.unconfirmedPDU.unconfirmedService.present = UnconfirmedService_PR_informationReport;

	InformationReport_t* report =  &(mmsPdu->choice.unconfirmedPDU.unconfirmedService.choice.informationReport);
	report->variableAccessSpecification.present = VariableAccessSpecification_PR_variableListName;

	if (domainId != NULL) {
		report->variableAccessSpecification.choice.variableListName.present = ObjectName_PR_domainspecific;
		report->variableAccessSpecification.choice.variableListName.choice.domainspecific.domainId.size = strlen(domainId);
		report->variableAccessSpecification.choice.variableListName.choice.domainspecific.domainId.buf = domainId;
		report->variableAccessSpecification.choice.variableListName.choice.domainspecific.itemId.size = strlen(itemId);
		report->variableAccessSpecification.choice.variableListName.choice.domainspecific.itemId.buf = itemId;
	}
	else {
		report->variableAccessSpecification.choice.variableListName.present = ObjectName_PR_vmdspecific;
		report->variableAccessSpecification.choice.variableListName.choice.vmdspecific.size = strlen(itemId);
		report->variableAccessSpecification.choice.variableListName.choice.vmdspecific.buf = itemId;
	}

	int variableCount = LinkedList_size(values);

	report->listOfAccessResult.list.array = calloc(variableCount, sizeof(AccessResult_t*));
	report->listOfAccessResult.list.size = variableCount;
	report->listOfAccessResult.list.count = variableCount;

	/* iterate values list and add values to the accessResultList */
	LinkedList value = LinkedList_getNext(values);

	int i;

	for (i = 0; i < variableCount; i++) {
		report->listOfAccessResult.list.array[i] = calloc(1, sizeof(AccessResult_t));

		AccessResult_t* accessResult = report->listOfAccessResult.list.array[i];

		mmsMsg_addResultToResultList(accessResult, (MmsValue*) value->data);

		value = LinkedList_getNext(value);
	}

	if (DEBUG) printf("sendInfReport: %i items\n", variableCount);

	ByteBuffer* reportBuffer = ByteBuffer_create(NULL, self->maxPduSize);

	asn_enc_rval_t rval =
			der_encode_to_buffer(&asn_DEF_MmsPdu, mmsPdu, reportBuffer->buffer, self->maxPduSize);

	if (DEBUG) printf("sendInfReport: %i encoded\n", rval.encoded);

	if (DEBUG) xer_fprint(stdout, &asn_DEF_MmsPdu, mmsPdu);

	reportBuffer->size = rval.encoded;

	IsoConnection_sendMessage(self->isoConnection, reportBuffer);

	freeMmsInformationReportPdu(mmsPdu);

	ByteBuffer_destroy(reportBuffer);
}
#endif
